/* Helper functions for the C port of Watchperson
 *
 * By John Elliott, 2021-11-02
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <dos.h>
#include "screen.h"

static byte far *framebuf = NULL;

/* Pixel-double a byte to a word */
static word stretch[256] = 
{
	0x0000, 0x0003, 0x000C, 0x000F, 0x0030, 0x0033, 0x003C, 0x003F,
	0x00C0, 0x00C3, 0x00CC, 0x00CF, 0x00F0, 0x00F3, 0x00FC, 0x00FF,
	0x0300, 0x0303, 0x030C, 0x030F, 0x0330, 0x0333, 0x033C, 0x033F,
	0x03C0, 0x03C3, 0x03CC, 0x03CF, 0x03F0, 0x03F3, 0x03FC, 0x03FF,
	0x0C00, 0x0C03, 0x0C0C, 0x0C0F, 0x0C30, 0x0C33, 0x0C3C, 0x0C3F,
	0x0CC0, 0x0CC3, 0x0CCC, 0x0CCF, 0x0CF0, 0x0CF3, 0x0CFC, 0x0CFF,
	0x0F00, 0x0F03, 0x0F0C, 0x0F0F, 0x0F30, 0x0F33, 0x0F3C, 0x0F3F,
	0x0FC0, 0x0FC3, 0x0FCC, 0x0FCF, 0x0FF0, 0x0FF3, 0x0FFC, 0x0FFF,
	0x3000, 0x3003, 0x300C, 0x300F, 0x3030, 0x3033, 0x303C, 0x303F,
	0x30C0, 0x30C3, 0x30CC, 0x30CF, 0x30F0, 0x30F3, 0x30FC, 0x30FF,
	0x3300, 0x3303, 0x330C, 0x330F, 0x3330, 0x3333, 0x333C, 0x333F,
	0x33C0, 0x33C3, 0x33CC, 0x33CF, 0x33F0, 0x33F3, 0x33FC, 0x33FF,
	0x3C00, 0x3C03, 0x3C0C, 0x3C0F, 0x3C30, 0x3C33, 0x3C3C, 0x3C3F,
	0x3CC0, 0x3CC3, 0x3CCC, 0x3CCF, 0x3CF0, 0x3CF3, 0x3CFC, 0x3CFF,
	0x3F00, 0x3F03, 0x3F0C, 0x3F0F, 0x3F30, 0x3F33, 0x3F3C, 0x3F3F,
	0x3FC0, 0x3FC3, 0x3FCC, 0x3FCF, 0x3FF0, 0x3FF3, 0x3FFC, 0x3FFF,
	0xC000, 0xC003, 0xC00C, 0xC00F, 0xC030, 0xC033, 0xC03C, 0xC03F,
	0xC0C0, 0xC0C3, 0xC0CC, 0xC0CF, 0xC0F0, 0xC0F3, 0xC0FC, 0xC0FF,
	0xC300, 0xC303, 0xC30C, 0xC30F, 0xC330, 0xC333, 0xC33C, 0xC33F,
	0xC3C0, 0xC3C3, 0xC3CC, 0xC3CF, 0xC3F0, 0xC3F3, 0xC3FC, 0xC3FF,
	0xCC00, 0xCC03, 0xCC0C, 0xCC0F, 0xCC30, 0xCC33, 0xCC3C, 0xCC3F,
	0xCCC0, 0xCCC3, 0xCCCC, 0xCCCF, 0xCCF0, 0xCCF3, 0xCCFC, 0xCCFF,
	0xCF00, 0xCF03, 0xCF0C, 0xCF0F, 0xCF30, 0xCF33, 0xCF3C, 0xCF3F,
	0xCFC0, 0xCFC3, 0xCFCC, 0xCFCF, 0xCFF0, 0xCFF3, 0xCFFC, 0xCFFF,
	0xF000, 0xF003, 0xF00C, 0xF00F, 0xF030, 0xF033, 0xF03C, 0xF03F,
	0xF0C0, 0xF0C3, 0xF0CC, 0xF0CF, 0xF0F0, 0xF0F3, 0xF0FC, 0xF0FF,
	0xF300, 0xF303, 0xF30C, 0xF30F, 0xF330, 0xF333, 0xF33C, 0xF33F,
	0xF3C0, 0xF3C3, 0xF3CC, 0xF3CF, 0xF3F0, 0xF3F3, 0xF3FC, 0xF3FF,
	0xFC00, 0xFC03, 0xFC0C, 0xFC0F, 0xFC30, 0xFC33, 0xFC3C, 0xFC3F,
	0xFCC0, 0xFCC3, 0xFCCC, 0xFCCF, 0xFCF0, 0xFCF3, 0xFCFC, 0xFCFF,
	0xFF00, 0xFF03, 0xFF0C, 0xFF0F, 0xFF30, 0xFF33, 0xFF3C, 0xFF3F,
	0xFFC0, 0xFFC3, 0xFFCC, 0xFFCF, 0xFFF0, 0xFFF3, 0xFFFC, 0xFFFF,

};

/* Get current video mode */
int getmode(void)
{
	union REGS rg;

	rg.h.ah = 0x0F;
	int86(0x10, &rg, &rg);
	return rg.h.al;
}


/* Request a video mode change */
void setmode (byte m)
{
	union REGS rg;

	rg.x.ax = m;
	int86(0x10, &rg, &rg);
}

/* After a mode change, select the appropriate palette and (if need be) hide
 * the cursor */
void initscr()
{
	union REGS rg;
	framebuf = MK_FP(((scrmode == 7) ? 0xB000 : 0xB800),0);

	if (scrmode == 4 || scrmode == 5)	/* Select palette */
	{
		rg.h.ah = 0x0B;
		rg.x.bx = 0x100;
		int86(0x10, &rg, &rg);
		rg.h.ah = 0x0B;
		rg.x.bx = 0x10;
		int86(0x10, &rg, &rg);

	}
	else if (scrmode == 7)
	{
		rg.h.ah = 1;	/* Hide cursor */
		rg.x.cx = 0x1F1E;
		int86(0x10, &rg, &rg);
	}
}


/* Get the system tick count. This uses the DOS 'get time' call, and returns
 * the number of centiseconds; in practice the number will increase in steps
 * of 5 or so */
int get_ticks(void)
{
	union REGS rg;

	rg.h.ah = 0x2C;
	int86(0x21, &rg, &rg);

	return rg.h.dh * 100 + rg.h.dl;
}


/* Attributes for MDA screen:
	YELLOW => normal
	RED    => inverse
	GREEN  => bright
	3      => blinking (not used) */	
static byte mda_attr[4] = { 0x07, 0x70, 0x0F, 0x87 };

/* Attribute bitmasks for CGA screen: YELLOW, RED, GREEN, unused */
static word cga_attr[4] = { 0xFFFF, 0xAAAA, 0x5555, 0xA5A5 };

/* Paint a single character */
void show_char(int row, int col, ATTR attr, byte c)
{
	char txt[2];

	txt[0] = c;
	txt[1] = 0;
	show_msg(row, col, attr, txt);
}

/* Paint a string */
void show_msg(int row, int col, ATTR attr, const char *txt)
{
	unsigned offset;
	byte *ch;
	word bits;
	word mask;
	word step;
	int y;

	/* In MDA mode, just poke the characters and attributes into the 
	 * framebuffer */
	if (scrmode == 7)
	{
		offset = row * 160 + col * 2;

		while (*txt)
		{
			framebuf[offset++] = *txt;
			framebuf[offset++] = mda_attr[attr & 3];
			++txt;
		}	
		return;
	}
	/* In CGA mode, each character has to be drawn as a bitmap */
	offset = row * 320 + col * 2;
	mask = cga_attr[attr & 3];
	while (*txt)
	{
		ch = font + 8 * ((byte)txt[0]);
		step = 0;
		for (y = 0; y < 8; y++)
		{
			bits = stretch[ch[y]] & mask;
			framebuf[offset + step    ] = (bits >> 8);
			framebuf[offset + step + 1] = bits & 0xFF;

			if (y & 1) step = (step & 0x1FFF) + 80;
			else step = step | 0x2000;
		}
		offset += 2;
		txt++;
	}
}

