/*

	*** WATCHPERSON ***

	(C) MAC OGLESBY   MARCH 1980

	LAST REVISED: 2021-11-02

	COPYRIGHT 1980 BY MAC OGLESBY

	C / PCDOS PORT BY JOHN ELLIOTT
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#define EXT
#include "screen.h"
#include <conio.h>

#define GREY 0xB1	/* Codepage 437 "50% grey" character */



typedef enum 
{
	NORTH = -40, EAST = 1, SOUTH = 40, WEST = -1, NONE = 0
} NSEW;

typedef const char *PSTR;

const PSTR directions[4] = { " N ", " E ", " S ", " W " };
NSEW offset[4] = { NORTH, EAST, SOUTH, WEST };

/* Characters to use for footprints in the N/E/S/W directions. On 
 * CGA these are the last 4 characters of the character set. On 
 * MDA they will all be set to 0xFE. */

unsigned char footprints[] = { 0xFC, 0xFD, 0xFE, 0xFF };

/* Use a C structure to replace the DATA lines in the BASIC program */
typedef struct village
{
	const char *name;	/* Full name */
	const char *short1;	/* First half of split name */
	const char *short2;	/* Second half of split name */
	int width;		/* Village width */
	int height;		/* Village height */
	int nstart;		/* Number of start points */
	int nstreets;		/* Number of streets */
	int totsteps;		/* Number of steps in total */
	int startpx;		/* Location to start drawing */
	int startpos[25];	/* Locations of startpoints */
	int street[50];		/* Street drawing directions / lengths */

} VILLAGE;

const VILLAGE villages[3] = 
{
	{ "Ogleboro", "Ogle-", "boro", 22, 17,  14, 14,  120, 171,
		{ 56, 65, 171, 176, 185, 190, 336, 345, 350, 496, 505, 
		  510, 611, 625 },
		{ EAST, 19,  SOUTH, 8,  WEST, 5,  SOUTH, 3,  WEST, 14, 
		  NORTH, 11, EAST, 5,  SOUTH, 8, EAST, 14, NORTH, 4,  
		  WEST, 14, NORTH, 7, EAST, 9, SOUTH, 10 }
	},

	{ "Kunktown", "Kunk-", "town", 24, 19,  19, 14,  157, 170,
		{ 54, 63, 67, 170, 174, 183, 187, 191, 334, 343, 347, 454, 463,
		  467, 570, 574, 583, 694, 711 },
		{ EAST, 21, SOUTH, 13, WEST, 17, NORTH, 6, EAST, 13, NORTH, 10,
		  WEST, 13, SOUTH, 7, EAST, 13, NORTH, 7, WEST, 4, SOUTH, 13, 
		  WEST, 13, NORTH, 10 }
	},	

	{ "Lindburg", "Lind-", "burg", 28, 20, 25, 20, 195, 328,
		{ 56, 65, 69, 73, 172, 176, 185, 189, 328, 332, 336, 345, 349,
		  492, 496, 505, 509, 619, 616, 625, 629, 728, 732, 736, 753 },
		{ EAST, 21, SOUTH, 7, WEST, 17, SOUTH, 3, WEST, 4, NORTH, 10,
		  EAST, 4, SOUTH, 4, EAST, 17, SOUTH, 3, WEST, 4, NORTH, 14,
		  WEST, 9, SOUTH, 17, EAST, 17, NORTH, 17, WEST, 4, SOUTH, 3,
		  WEST, 17, SOUTH, 4 }
	}
};

char pet_screen[880];

VILLAGE cur_village;


void setup_village(int v)
{
	int row, col, px, i, j;
	char line[41];

	cur_village = villages[v];	

	/* Set up the screen */
	memset(pet_screen, ' ', sizeof(pet_screen));
	col = ((40 - cur_village.width) / 2) + 1;
	for (row = 0; row < cur_village.height; row++)
	{
		memset(pet_screen + 40*row + col, GREY, cur_village.width);
	}
	/* Draw the streets */
	px = cur_village.startpx;
	pet_screen[px] = ' ';
	for (i = 0; i < 2*cur_village.nstreets; i += 2)
	{
		for (j = 0; j < cur_village.street[i+1]; j++)
		{
			px += cur_village.street[i];
			pet_screen[px] = ' ';
		}
	}
	/* Draw the start points */
	for (j = 0; j < cur_village.nstart; j++)
	{
		pet_screen[cur_village.startpos[j]] = 'A' + j;
	}

	/* Draw the compass */
	pet_screen[  2] = 'N';
	pet_screen[ 42] = '|';
	memcpy(pet_screen + 80, "W-+-E", 5);
	pet_screen[ 82] = 0x04;
	pet_screen[122] = '|';
	pet_screen[162] = 'S';

	/* Draw the map title */
	memcpy(pet_screen + 280, "Map", 3);
	memcpy(pet_screen + 360, "of", 2);
	memcpy(pet_screen + 440, cur_village.short1, strlen(cur_village.short1));
	memcpy(pet_screen + 520, cur_village.short2, strlen(cur_village.short2));

	/* Paint the PET screen to the real screen */	
	for (row = 0; row < 22; row++)
	{
		sprintf(line, "%-40.40s", pet_screen + 40 * row);
		show_msg(row, 0, YELLOW, line);
	}	
}


/* Show a prompt on the bottom line, wait for input, uppercase it and return */
char prompt(const char *p)
{
	ATTR attr = RED;
	int t = get_ticks();
	int flash = 0;
	char c;

	/* Swallow keyboard input */
	while (kbhit())
	{
		getch();
	}
	show_msg(23, 0, attr, p);
	while (1)
	{
		if (kbhit())
		{
			c = getch();
			if (islower(c)) c = toupper(c);
			show_msg(23, 0, YELLOW, "                                        ");
			return c;
		}
		if (t != get_ticks())
		{
			++flash;
			t = get_ticks();
		}
		if (flash >= 20)
		{
			attr ^= (RED ^ GREEN);
			flash = 0;
			show_msg(23, 0, attr, p);
		}

	}
	return 0;
}


/* Wait for the specified number of system clock ticks */
void mpause(int n)
{
	int t = get_ticks();

	for (; n >= 0; n--)
	{
		while (t == get_ticks()) {}
		t = get_ticks();
	}
}

/* Repaint a cell on the map */
void update_map(int offset, byte c)
{
	/* Draw footsteps in green, anything else in yellow */
	ATTR a = (c >= 0xFC) ? GREEN : YELLOW;

	pet_screen[offset] = c;
	
	show_char(offset / 40, offset % 40, a, c);
}


/* Find the appropriate footprint graphic for a given direction of travel */
char footprint(NSEW d)
{
	switch(d)
	{
		case NORTH: return footprints[0];
		case EAST:  return footprints[1];
		case SOUTH: return footprints[2];
		case WEST:  return footprints[3];
	}
	return '?';
}

/* Main game loop. This has been converted from the BASIC without much 
 * attempt to divide it into functions, so you will find 'goto' and other
 * horrors within */
int watchperson(void)
{
	char want_ins, sel_v, startpos, dir, again;
	int vn, j, z0, cr;
	char pbuf[41];
	int steps;
	NSEW d, d1;
	char fp;
	byte cell;

	setmode(scrmode);
	initscr();
	/* Draw Ogleboro to serve as a title screen */	
	setup_village(0);

	show_msg(19,0, GREEN, "Welcome, Watchperson...");
	do
	{
		want_ins = prompt("Want instructions?  Yes or No");
	}
	while (want_ins != 'Y' && want_ins != 'N');
	setmode(scrmode);
	initscr();
			     /*1...5...10...15...20...25...30...35...40 */
	show_msg(0, 0, GREEN, "As watchperson, you have been hired to");
	show_msg(1, 0, GREEN, "walk all the streets of a village ");
	show_msg(2, 0, GREEN, "without retracing any of your steps.");
	show_msg(4, 0, GREEN, "Choose a village:  ");
	show_msg(4, 20,RED, "O");
	show_msg(4, 21, GREEN, "gleboro, ");
	show_msg(4, 30, RED, "K");
	show_msg(4, 31, GREEN, "unktown,");
	show_msg(5, 0, GREEN, "or ");
	show_msg(5, 3, RED, "L");
	show_msg(5, 4, GREEN, "indburg. Lindburg is the hardest.");
	do
	{
		sel_v = prompt("Please type  O, K or L");

	} while (sel_v != 'O' && sel_v != 'K' && sel_v != 'L');

	switch (sel_v)
	{
		case 'O': vn = 0; break;
		case 'K': vn = 1; break;
		case 'L': vn = 2; break;
	}
	if (want_ins != 'N')
	{
		int n;
/* Draw the instructions screen if required */
		setmode(scrmode);
		initscr();
		for (n = 0; n < strlen("Footprints"); n++)
		{
			show_char(4, 20 + n, YELLOW, "Footprints"[n]);
			mpause(1);
		}
		show_char(6, 18, GREEN, footprints[3]);
		for (n = 1; n <= 12; n++)
		{
			mpause(2);
			show_char(6, 18 - n, GREEN, footprints[3]);
		}
		for (n = 1; n <= 4; n++)
		{
			mpause(2);
			show_char(6 - n, 6, GREEN, footprints[0]);	
		}
		for (n = 1; n <= 9; n++)
		{
			mpause(2);
			show_char(2, 6 + n, GREEN, footprints[1]);	
		}
		for (n = 1; n <= 6; n++)
		{
			mpause(2);
			show_char(2 + n, 15, GREEN, footprints[2]);	
		}	
		for (n = 1; n <= 12; n++)
		{
			mpause(2);
			show_char(8, 15 - n, GREEN, footprints[3]);
		}
		for (n = 13; n <= 15; n++)
		{
			mpause(2);
			show_char(8, 15 - n, GREEN, '.');
		}
		mpause(16);
		show_msg(10, 0, YELLOW, "Although you may ");
		show_msg(10, 17, RED, "not");
		show_msg(10, 21, YELLOW, "walk twice on");
		show_msg(10, 35, RED, "any");
		show_msg(11, 0, YELLOW, "street, you may revisit corners.");
		show_msg(13, 0, YELLOW, "Remember, your task is to walk along");
		show_msg(14, 0, RED, "all");
		show_msg(14, 4, YELLOW, "of the streets.");
		show_msg(16, 0, GREEN, "Good luck!");

		prompt("Press RETURN to start...");
	}
	setup_village(vn);

/* Choose start point */
restart:
	z0 = 0;
	steps = 0;
	for (j = 0; j < cur_village.nstart; j++)
	{
		update_map(cur_village.startpos[j], 'A' + j);
	}
	sprintf(pbuf, "%-40.40s", "Start where?");
	show_msg(22, 0, GREEN, pbuf);

	steps = 0;
	startpos = 0;	
	while (startpos < 'A' || startpos >= 'A' + cur_village.nstart)
	{
		startpos = prompt("Pick one of the letters shown");	
	}
	z0 = cur_village.startpos[startpos - 'A'];
/* Main game loop */
	while (1)
	{
		strcpy(pbuf, "Walk which way?");	
/* Find suitable directions */
		for (j = 0; j < 4; j++)
		{
			if (pet_screen[z0 + offset[j]] == ' ') 
				strcat(pbuf, directions[j]);
		}
		/* Pad to 40 chars */
		for (j = strlen(pbuf); j < 40; j++) pbuf[j] = ' ';
		pbuf[40] = 0;
	
		show_msg(22, 0, GREEN, pbuf);
		if (steps == 0)
		{
			dir = prompt("Type X to cancel starting point");
			if (dir == 'X') goto restart;
		}
		else
		{
			dir = prompt(" ");
		}
		switch(dir)
		{
			case 'N': d = NORTH; break;
			case 'S': d = SOUTH; break;
			case 'E': d = EAST; break;
			case 'W': d = WEST; break;
			default:  d = NONE; break;
		}
		fp = footprint(d);
		if (pet_screen[z0 + d] != ' ') continue;
		if (steps == 0)
		{
/* On the first move, remove starting point indicators */
			for (j = 0; j < cur_village.nstart; j++)
			{
				update_map(cur_village.startpos[j], ' ');
			}
		}
		++steps;
		update_map(z0, fp);
stepping:
		cr = 0;
stepping1:
		z0 += d;
		update_map(z0, fp);
		++steps;
		for (j = 0; j < 4; j++)
		{
			if (pet_screen[z0 + offset[j]] == ' ')
			{
				++cr;
				d1 = offset[j];
			}
		}
		switch(cr)
		{
			case 1:
			case 11:	/* Only one direction possible */
				d = d1;
				fp = footprint(d);
				goto stepping;
			case 10:	/* End of game */
				goto endgame;
			case 2:		/* Choice of routes */
			case 3:
			case 4:
			case 12:
				d = NONE;
				continue;
			case 0: 	/* No route possible */
				cell = pet_screen[z0 + d];
				if (cell >= 0xFC)
				{
					cr = 10;
					goto stepping1;
				}	
				goto endgame;
		}

	} /* End of main game loop */	
endgame:
	if (steps != cur_village.totsteps)
	{
		sprintf(pbuf, "%-40.40s", "You were supposed to walk every street!");
	}
	else
	{
		sprintf(pbuf, "%-40.40s", "Excellent - you walked every street!");
	}
	for (j = 0; j < 10; j++)
	{
		show_msg(22, 0, (j & 1) ? RED : GREEN, pbuf);
		mpause(5);	
	}
	do
	{
		again = prompt("Press Y to play again, N to stop");
	} while (again != 'Y' && again != 'N');
	return (again == 'Y');
}



/* Main loop gets options, and then runs the game until the user answers
 * 'no' to "Play again?" */
int main(int argc, char **argv)
{
	int n;

	oldmode = getmode();

	scrmode = (oldmode == 7) ? 7 : 4;	

	for (n = 1; n < argc; n++)
	{
		if (argv[n][0] == '-') switch (argv[n][1])
		{
			case 'm': case 'M': scrmode = 7; break;
			case 'c': case 'C': scrmode = 4; break;
		}
	}
	if (scrmode == 7) memset(footprints, 0xFE, sizeof(footprints));
	setmode(scrmode);
	if (getmode() != scrmode)
	{
		setmode(oldmode);
		fprintf(stderr, "Cannot select video mode %d.\n", scrmode);
		return 1;
	}
	initscr();
	while (watchperson())
	{
	}	

	setmode(oldmode);
	printf("Thanks for walking my streets.\n"
		"   Bye for now, Watchperson...");
	return 0;
}
